#include <qpushbutton.h>
#include <qcombobox.h>
#include <qradiobutton.h>
#include <qapplication.h>
#include <qspinbox.h>
#include <qsortedlist.h>
#include <qlineedit.h>
#include <qlabel.h>
#include <qcheckbox.h>

#include <float.h>


#include "../../../Core/Scene/src/Scene.h"
#include "../../../Core/Scene/src/Object.h"
#include "../../../Core/Transformer/src/Transformer.h"
#include "../../../Core/DataStructures/src/NeighbourHood.h"
#include "../../../Core/DataStructures/src/SurfelCollection.h"
#include "../../../Core/CoreResourceTools/src/CoreResourceTools.h"
#include "../../../Core/CoreTools/src/CoreTools.h"
#include "../../../DataTypes/src/MyDataTypes.h"
#include "../../../Utilities/src/Matrix.h"
#include "../../../Utilities/src/MyMathTools.h"
#include "../../../Utilities/src/FileTools.h"
#include "../../../Utilities/src/StringTools.h"

#include "Bilateral_DialogImpl.h"


// ******
// global
// ******

// the single Bilateral_DialogImpl instance, only
// accesible by getInstance() and
// destroyInstance()
static Bilateral_DialogImpl *instance = 0;


// *******************
// private c'tor/d'tor
// *******************

Bilateral_DialogImpl::Bilateral_DialogImpl (QWidget *parent, const char *name, bool modal, WFlags flags)
    : Bilateral_Dialog (parent, name, modal, flags) {

	this->initDialog();
}

Bilateral_DialogImpl::~Bilateral_DialogImpl() {
    // no need to delete child widgets, Qt does it all for us
}


// **************
// public methods
// **************

Bilateral_DialogImpl *Bilateral_DialogImpl::getInstance() {

	if (instance == 0) {
		instance = new Bilateral_DialogImpl();
	}
	return instance;

}

void Bilateral_DialogImpl::destroyInstance() {

	if (instance != 0) {
		instance->close();
		delete instance;
	}
	instance = 0;

}

// ***************
// private methods
// ***************

void Bilateral_DialogImpl::initDialog () {
	
	defRelVariance = 0.5f;
	RelativeFilterVarianceLineEdit->setText (QString::number (defRelVariance));
	defRobustFactor = 1.f;
	RobustFilterFactorLineEdit->setText (QString::number (defRobustFactor));
}



// *************
// private slots
// *************


/* Perform weighted average filtering.
*/
void Bilateral_DialogImpl::handleWeightedAverageFiltering() {

	unsigned int i, j;
	unsigned int nOfNeighbors;
	unsigned int nSurfels;
	Scene *scene;
	SurfelCollection *objectSurfels;
	
	QVector<SurfelInterface> *objectSurfelVector;
	QVector<SurfelInterface> *filteredSurfelVector;

	Vector3D* surfelPositions;
	NeighbourHood *neighborHood;

	Vector3D newPosition, tmpVector;
	float sumWeights, w;

	float maxDist, d;
	float relVariance;
	bool isNumber;

	scene = Scene::getInstance();

	if (scene->getNofObjects() > 0) {

		QApplication::setOverrideCursor (Qt::waitCursor);

		// get the relative filter variance from the dialog
		relVariance = RelativeFilterVarianceLineEdit->text().toFloat (&isNumber);
		if (isNumber == false) {
			RelativeFilterVarianceLineEdit->setText (QString::number (defRelVariance));
			relVariance = defRelVariance;
		}

		// get vector of surfels
		objectSurfels = scene->getActiveObject()->getSurfelCollection();
		objectSurfelVector = objectSurfels->buildSurfelVector();
		nSurfels = objectSurfels->getNofSurfels();

		// get vector of surfel positions
		surfelPositions = new Vector3D[nSurfels];

		for (i = 0; i < nSurfels; i++) {
			surfelPositions[i] = objectSurfelVector->at(i)->getPosition();
		}

		// allocate vector of filtered surfels
		filteredSurfelVector = new QVector<SurfelInterface>(nSurfels);

		// set up neighborhood
		neighborHood = new NeighbourHood(surfelPositions, nSurfels);
		nOfNeighbors = this->AverageFilteringNumNeighborsSpinBox->value();
		// surfel itself is in neighborhood, so we increase nOfNeighbors by one
		nOfNeighbors += 1;
		neighborHood->setNofNeighbours (nOfNeighbors);

		// filter surfels
		for (i=0; i<nSurfels; i++) {

			neighborHood->setSourcePoint (surfelPositions[i]);

			newPosition = Vector3D(0.f, 0.f, 0.f);
			sumWeights = 0.f;

			// getNeighbourPosition() actually returns the position of the n-th nearest neighbor!
			tmpVector = surfelPositions[i] - neighborHood->getNeighbourPosition(nOfNeighbors-1);
			maxDist = tmpVector.getLength();

			for (j = 0; j < nOfNeighbors; j++) {
				// compute relative distance of current neighbor
				tmpVector = surfelPositions[i] - neighborHood->getNeighbourPosition(j);
				d = tmpVector.getLength() / maxDist;

				// filter weighted by relative distance
				w = exp(-0.5 * d * d / relVariance);

				newPosition += w * neighborHood->getNeighbourPosition(j);
				sumWeights += w;
			}

			newPosition /= sumWeights;

			// add a copy of the original surfel to the object surfel collection...
			filteredSurfelVector->insert(i, objectSurfels->addSurfel(objectSurfelVector->at(i), false));
			// ...and adjust its position
			scene->setSurfelPosition(filteredSurfelVector->at(i), newPosition, false);

		}

		// remove original surfels from object surfel collection
		for (i=0; i<nSurfels; i++) {
			objectSurfels->removeSurfel(objectSurfelVector->at(i), false);
		}

		scene->emitObjectModifiedSignal ();
		scene->emitRenderFrameSignal ();

		QApplication::restoreOverrideCursor ();

		delete objectSurfelVector;
		delete filteredSurfelVector;
		delete[] surfelPositions;
		delete neighborHood;
	}

}


/* Perform iterative projection.
*/
void Bilateral_DialogImpl::handleIterativeProjection() {

	unsigned int i, j;
	unsigned int nOfNeighbors;
	unsigned int nSurfels;
	Scene *scene;
	SurfelCollection *objectSurfels;
	
	QVector<SurfelInterface> *objectSurfelVector;
	QVector<SurfelInterface> *filteredSurfelVector;

	Vector3D* surfelPositions;
	NeighbourHood *neighborHood;

	float relVariance;
	float rFactor;
	float t1, t2;

	Vector3D newNormal, robustNormal;
	Vector3D wAvg, robustAvg;
	Vector3D axis1, axis2;
	Vector3D diff;
	Vector3D newPosition;

	bool isNumber;


	scene = Scene::getInstance();

	if (scene->getNofObjects() > 0) {

		QApplication::setOverrideCursor (Qt::waitCursor);

		// get the relative filter variance from the dialog
		relVariance = RelativeFilterVarianceLineEdit->text().toFloat (&isNumber);
		if (isNumber == false) {
			RelativeFilterVarianceLineEdit->setText (QString::number (defRelVariance));
			relVariance = defRelVariance;
		}

		// get the robust filter factor from the dialog
		rFactor = RobustFilterFactorLineEdit->text().toFloat (&isNumber);
		if (isNumber == false) {
			RobustFilterFactorLineEdit->setText (QString::number (defRobustFactor));
			rFactor = defRobustFactor;
		}

		// get vector of surfels
		objectSurfels = scene->getActiveObject()->getSurfelCollection();
		objectSurfelVector = objectSurfels->buildSurfelVector();
		nSurfels = objectSurfels->getNofSurfels();

		// get vector of surfel positions
		surfelPositions = new Vector3D[nSurfels];

		for (i = 0; i < nSurfels; i++) {
			surfelPositions[i] = objectSurfelVector->at(i)->getPosition();
		}

		// allocate vector of filtered surfels
		filteredSurfelVector = new QVector<SurfelInterface>(nSurfels);

		// set up neighborhood
		neighborHood = new NeighbourHood(surfelPositions, nSurfels);
		nOfNeighbors = this->AverageFilteringNumNeighborsSpinBox->value();
		// surfel itself is in neighborhood, so we increase nOfNeighbors by one
		nOfNeighbors += 1;
		neighborHood->setNofNeighbours (nOfNeighbors);

		// filter surfels
		for (i=0; i<nSurfels; i++) {

			neighborHood->setSourcePoint (surfelPositions[i]);

			// add a copy of the original surfel to the object surfel collection
			filteredSurfelVector->insert(i, objectSurfels->addSurfel(objectSurfelVector->at(i), false));

			// compute new normal and weighted average of weighted surfel neighborhood
			computeNormal(neighborHood, surfelPositions[i], nOfNeighbors, relVariance, newNormal, wAvg);
			if(Vector3D::dotProduct(newNormal, filteredSurfelVector->at(i)->getNormal()) < 0) {
				newNormal *= -1;
			}
			scene->setSurfelNormal(filteredSurfelVector->at(i), newNormal, false);

			// do one iteration with non-robust normal, non-robust average
			diff = surfelPositions[i] - wAvg;
			newNormal.getTwoOrthogonals(axis1, axis2);
			t1 = Vector3D::dotProduct(axis1, diff);
			t2 = Vector3D::dotProduct(axis2, diff);
			newPosition = wAvg + t1*axis1 + t2*axis2;


			// iterate with robust normals, robust averages
			for(j=0; j<10; j++) {
				// compute robust normal and weighted average
				computeRobustNormal(neighborHood, newPosition, nOfNeighbors, relVariance, newNormal, rFactor * (float)j/9.f, robustNormal, robustAvg);
				if(Vector3D::dotProduct(robustNormal, filteredSurfelVector->at(i)->getNormal()) < 0) {
					robustNormal *= -1;
				}

				diff = newPosition - robustAvg;
				robustNormal.getTwoOrthogonals(axis1, axis2);
				t1 = Vector3D::dotProduct(axis1, diff);
				t2 = Vector3D::dotProduct(axis2, diff);
				newPosition = robustAvg + t1*axis1 + t2*axis2;
				newNormal = robustNormal;
			}

			// adjust position
			scene->setSurfelPosition(filteredSurfelVector->at(i), newPosition, false);
		}

		// remove original surfels from object surfel collection
		for (i=0; i<nSurfels; i++) {
			objectSurfels->removeSurfel(objectSurfelVector->at(i), false);
		}

		scene->emitObjectModifiedSignal ();
		scene->emitRenderFrameSignal ();

		QApplication::restoreOverrideCursor ();

		delete objectSurfelVector;
		delete filteredSurfelVector;
		delete[] surfelPositions;
		delete neighborHood;
	}

}


/******************
 * Private methods.
 ******************/

// normal is not consistent!!!
void Bilateral_DialogImpl::computeNormal(NeighbourHood *neighborHood, Vector3D position, unsigned int nNeighbors, float relVariance, Vector3D& normal, Vector3D& wAvg) {

	unsigned int i;
	Vector3D diffVector;
	float w, sumW;
	float d, maxD, relD;

	Mgc::Eigen *eigenSolver	= new Mgc::Eigen(3);

	// compute sums of products, these values are actually the elements of the 3x3 covariance matrix
	Mgc::Real fSumXX = 0.0f, fSumXY = 0.0f, fSumXZ = 0.0f;
	Mgc::Real fSumYY = 0.0f, fSumYZ = 0.0f, fSumZZ = 0.0f;

	// set position where normal is going to be approximated
	neighborHood->setSourcePoint(position);

	// find maximum distance to center position in neighborhood
	diffVector = neighborHood->getNeighbourPosition(nNeighbors-1) - position;
	maxD = diffVector.getLength();

	// compute weighted average relative to center position
	wAvg.makeZero();
	sumW = 0.f;
	for (i = 0; i < nNeighbors; i++) {
		diffVector = neighborHood->getNeighbourPosition(i) - position;
		d = diffVector.getLength();
		relD = d / maxD;

		w = exp(-0.5 * relD*relD / relVariance);
		sumW += w;
		wAvg += w* neighborHood->getNeighbourPosition(i);

	}
	wAvg /= sumW;

	// compute weighted covariance matrix
	for (i = 0; i < nNeighbors; i++) {
		diffVector = wAvg - neighborHood->getNeighbourPosition(i);
		d = diffVector.getLength();
		relD = d / maxD;

		w = exp(-0.5 * relD*relD / relVariance);

		// add weighted contribution of current point to covariance matrix
		fSumXX += w*w * diffVector[0] * diffVector[0];
		fSumXY += w*w * diffVector[0] * diffVector[1];
		fSumXZ += w*w * diffVector[0] * diffVector[2];
		fSumYY += w*w * diffVector[1] * diffVector[1];
		fSumYZ += w*w * diffVector[1] * diffVector[2];
		fSumZZ += w*w * diffVector[2] * diffVector[2];
	}
	
	// setup the eigensolver
	eigenSolver->Matrix(0,0) = fSumXX;
	eigenSolver->Matrix(0,1) = fSumXY;
	eigenSolver->Matrix(0,2) = fSumXZ;
	eigenSolver->Matrix(1,0) = eigenSolver->Matrix(0,1);
	eigenSolver->Matrix(1,1) = fSumYY;
	eigenSolver->Matrix(1,2) = fSumYZ;
	eigenSolver->Matrix(2,0) = eigenSolver->Matrix(0,2);
	eigenSolver->Matrix(2,1) = eigenSolver->Matrix(1,2);
	eigenSolver->Matrix(2,2) = fSumZZ;
	
	// compute eigenstuff, smallest eigenvalue is in last position
	eigenSolver->DecrSortEigenStuff3();	
	
	// unit-length direction for best-fit line
	normal = Vector3D (eigenSolver->GetEigenvector(0,2), eigenSolver->GetEigenvector(1,2), eigenSolver->GetEigenvector(2,2));

	delete eigenSolver;
}


void Bilateral_DialogImpl::computeRobustNormal(NeighbourHood *neighborHood, Vector3D position, unsigned int nNeighbors, float relVariance, Vector3D initialNormal, float rFactor, Vector3D& normal, Vector3D& wAvg) {

	Vector3D axis1, axis2;

	initialNormal.normalize();
	initialNormal.getTwoOrthogonals(axis1, axis2);

	float trafo[9], invTrafo[9];
	float covar[9], tmp[9];

	unsigned int i;

	// transformation from local tangent space to object space
	// NOTE: the matrix has column major indexing
	trafo[0] = axis1[0];
	trafo[1] = axis1[1];
	trafo[2] = axis1[2];
	trafo[3] = axis2[0];
	trafo[4] = axis2[1];
	trafo[5] = axis2[2];
	trafo[6] = initialNormal[0];
	trafo[7] = initialNormal[1];
	trafo[8] = initialNormal[2];

	// get trafo from object space to local tangent space
	// NOTE: since trafo is orthonormal, we just take its transpose to get the invTrafo
	MtrTranspose3x3f(trafo, invTrafo);

	// compute (covariance matrix in local coordinates) * (inverse transformation)
	// NOTE: local covariance consists of
	// - uniform scaling specified as the "relative variance" in the neighborhood
	// - scale factor for robust filtering along normal direction
	float rFactorSquared = rFactor*rFactor;
	for(i=0; i<9; i++) {
		tmp[i] = invTrafo[i] / relVariance;
	}	
	tmp[2] *= rFactorSquared;
	tmp[5] *= rFactorSquared;
	tmp[8] *= rFactorSquared;

	// compute transpose(invTrafo) * ((covariance in local coords) * invTrafo) to get covariance 
	// matrix for robust filter
	// NOTE: transpose(invTrafo) = trafo
	MtrMult3x3f(trafo, tmp, covar);

	Vector3D diffVector;
	float w, sumW;
	float maxD, dSquared;
	float t1[3], t2[3];

	Mgc::Eigen *eigenSolver	= new Mgc::Eigen(3);

	// compute sums of products, these values are actually the elements of the 3x3 covariance matrix
	Mgc::Real fSumXX = 0.0f, fSumXY = 0.0f, fSumXZ = 0.0f;
	Mgc::Real fSumYY = 0.0f, fSumYZ = 0.0f, fSumZZ = 0.0f;

	// set position where normal is going to be approximated
	neighborHood->setSourcePoint(position);

	// find maximum distance to center position in neighborhood
	diffVector = neighborHood->getNeighbourPosition(nNeighbors-1) - position;
	maxD = diffVector.getLength();

	// compute weighted average relative to center position
	wAvg.makeZero();
	sumW = 0.f;
	for (i = 0; i < nNeighbors; i++) {
		diffVector = neighborHood->getNeighbourPosition(i) - position;
		diffVector /= maxD;

		// compute transpose(diffVector)*covar*diffVector
		t1[0] = diffVector[0];
		t1[1] = diffVector[1];
		t1[2] = diffVector[2];
		MtrMultVector3f(covar, t1, t2);
		dSquared = t1[0]*t2[0] + t1[1]*t2[1] + t1[2]*t2[2];
		
		// evaluate robust filter weight
		w = exp(-0.5 * dSquared);

		sumW += w;
		wAvg += w* neighborHood->getNeighbourPosition(i);

	}
	wAvg /= sumW;

	// compute weighted covariance matrix
	for (i = 0; i < nNeighbors; i++) {
		diffVector = wAvg - neighborHood->getNeighbourPosition(i);
		diffVector /= maxD;

		// compute transpose(diffVector)*covar*diffVector
		t1[0] = diffVector[0];
		t1[1] = diffVector[1];
		t1[2] = diffVector[2];
		MtrMultVector3f(covar, t1, t2);
		dSquared = t1[0]*t2[0] + t1[1]*t2[1] + t1[2]*t2[2];

		// evaluate robust filter weight
		w = exp(-0.5 * dSquared);

		// add weighted contribution of current point to covariance matrix
		fSumXX += w*w * diffVector[0] * diffVector[0];
		fSumXY += w*w * diffVector[0] * diffVector[1];
		fSumXZ += w*w * diffVector[0] * diffVector[2];
		fSumYY += w*w * diffVector[1] * diffVector[1];
		fSumYZ += w*w * diffVector[1] * diffVector[2];
		fSumZZ += w*w * diffVector[2] * diffVector[2];
	}
	
	// setup the eigensolver
	eigenSolver->Matrix(0,0) = fSumXX;
	eigenSolver->Matrix(0,1) = fSumXY;
	eigenSolver->Matrix(0,2) = fSumXZ;
	eigenSolver->Matrix(1,0) = eigenSolver->Matrix(0,1);
	eigenSolver->Matrix(1,1) = fSumYY;
	eigenSolver->Matrix(1,2) = fSumYZ;
	eigenSolver->Matrix(2,0) = eigenSolver->Matrix(0,2);
	eigenSolver->Matrix(2,1) = eigenSolver->Matrix(1,2);
	eigenSolver->Matrix(2,2) = fSumZZ;
	
	// compute eigenstuff, smallest eigenvalue is in last position
	eigenSolver->DecrSortEigenStuff3();	
	
	// unit-length direction for best-fit line
	normal = Vector3D (eigenSolver->GetEigenvector(0,2), eigenSolver->GetEigenvector(1,2), eigenSolver->GetEigenvector(2,2));

	delete eigenSolver;

}

